""" often used constants go here """

FDTD_X: int = 0
""" x-index (useful for indexing) """

FDTD_Y: int = 1
""" y-index (useful for indexing) """

FDTD_Z: int = 2
""" z-index (useful for indexing) """
